import io
import json
import uuid
from datetime import datetime, timedelta, timezone
from typing import Any, Generator

import numpy as np
import pandas as pd
from pydruid.utils.aggregators import cardinality
from pydruid.utils.filters import Bound, Dimension, Filter
from scipy import stats

import openapi.models.base_anomaly_event_detail  # needed to avoid circular import for any subclasses  # noqa: F401
from common.utils import logging
from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.common_model import DetectorType, Interval
from common.utils.detector.core import BatchDetector
from common.utils.event.core import EventService
from common.utils.event.data import (
    Compute,
    ComputeType,
    DataFlow,
    EventInfo,
    EventKey,
    EventMetadata,
    IpMask,
    JsonValue,
    ListString,
    MergePlan,
    MergePolicy,
    Score,
    Tactic,
    Technique,
)
from common.utils.flow.core import FlowService
from common.utils.flow.data import FlowGroupByQuery
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    DetectorRunStatus,
    DetectorStatusLabel,
    ExclusionLabel,
    ExclusionTypeInfo,
    UserConfigInfo,
)
from common.utils.vizfilter.core import (
    CompositeVizFilter,
    Condition,
    Criteria,
    VizFilter,
    VizFilterOption,
)
from openapi.models.port_scan_vertical_event_detail import PortScanVerticalEventDetail
from openapi.models.scanned_vm_details import ScannedVmDetails
from openapi.models.series import Series
from verticalportscan.detector.config import Config
from vmware.nsx.intelligence.nta.v1 import nta_pb2


class VerticalPortScanDetectionError(Exception):
    pass


class VerticalPortScan(BatchDetector):
    logger = logging.get_common_logger(__name__)

    def __init__(
        self,
        baseline_service: BaselineService,
        checkpoint_service: CheckpointService,
        user_config_service: UserConfigService,
        event_service: EventService,
        flow_service: FlowService,
        lookup_service: LookupService,
        now: datetime,
    ):
        super().__init__(baseline_service, checkpoint_service, user_config_service, event_service, now)
        self._flow_service = flow_service
        self._lookup_service = lookup_service
        self._now = now
        self._beginning_of_today = datetime.combine(now.date(), datetime.min.time())

    @staticmethod
    def detector_name() -> DetectorType:
        return DetectorType.VERTICAL_PORT_SCAN

    @staticmethod
    def _baseline_max_retention_time() -> timedelta:
        # need some check of baseline size, and dynamically adjust retention time to account
        # BUT need to ensure the same timedelta will be returned throughout the run
        return timedelta(days=float(Config.BASELINE_MAX_RETENTION_DAYS))

    def _generate_baseline(self, site_id: str, interval: Interval) -> PartialBaseline:
        VerticalPortScan.logger.info(f"generating vertical port scan baseline for interval {interval}")

        query_result = self._execute_port_scan_query([interval], site_id, Config.BASELINE_FLOW_QUERY_GRANULARITY)
        VerticalPortScan.logger.info(
            f"got {len(query_result.index)} rows from druid when building baseline for {interval}"
        )
        if len(query_result.index) != 0:
            query_result["timestamp"] = pd.to_datetime(query_result["timestamp"], format="%Y-%m-%d %H:%M:%S")
        buffer = io.BytesIO()
        query_result.to_parquet(buffer)
        return PartialBaseline(interval, buffer)

    def _get_baseline_intervals(self, missing_intervals: list[Interval]) -> list[Interval]:
        return BaselineService.convert_to_daily_intervals_with_trim(
            missing_intervals,
            self._beginning_of_today,
            timedelta(days=float(Config.BASELINE_MAX_HISTORICAL_CREATION_DAYS)),
        )

    def _execute_port_scan_query(
        self,
        intervals: list[Interval],
        site_id: str,
        granularity: str,
        included_computes: list[str] = [],
        excluded_computes: list[str] = [],
    ) -> pd.DataFrame:
        filters = [
            Dimension("algFlowType") != "DATA",
            Dimension("dstComputeType") == "VM",
            Dimension("dstSiteId") == site_id,
            Dimension("isDstInfoPresent") == True,  # noqa
        ]
        if included_computes != []:
            filters.append(Filter(type="in", dimension="dstVmId", values=included_computes))
        if excluded_computes != []:
            filters.append(~Filter(type="in", dimension="dstVmId", values=excluded_computes))

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=granularity,
            intervals=intervals,
            aggregations={"cardinalityCount": cardinality("dstPort")},
            dimensions=["dstVmId"],
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query)

    def _combine_baseline_objects(self, baselines: list[PartialBaseline]) -> io.BytesIO | None:
        """Take in partial baselines and combine into a stats table for detection

        result, partial baseline:
        <timestamp>:<timestamp>-baseline
            <timestamp>, vm1, 20
            <timestamp>, vm2, 10
            <timestamp>, vm3, 15
            <timestamp>, vm4, 30
            <timestamp>, vm5, 12
        <timestamp>:<timestamp>-baseline
            <timestamp>, vm1, 20
            <timestamp>, vm2, 10
            <timestamp>, vm3, 15
            <timestamp>, vm4, 30
            <timestamp>, vm5, 12
        ...

        combined baseline:
        vm1, <mean>, <std>, <days since first record>, <list of datetime of outliers>
        vm2, <mean>, <std>, <days since first record>, <list of datetime of outliers>
        vm3, <mean>, <std>, <days since first record>, <list of datetime of outliers>
        vm4, <mean>, <std>, <days since first record>, <list of datetime of outliers>
        vm5, <mean>, <std>, <days since first record>, <list of datetime of outliers>

        objects into
        meanMap             -> for each VM, calc mean and std, remove any datapoints mean +/- std*sigma, calc mean again
        stdDevMap           -> for each VM, calc mean and std, remove any datapoints mean +/- std*sigma, calc std again
        durationMap         -> for each VM, calc timedelta now - beginTime
        outlierDateTimeMap  -> for each VM, get a list of datetime of the outliers
        """
        if len(baselines) == 0:
            VerticalPortScan.logger.warn(
                "got empty list of baselines to combine, returning empty baseline for detection"
            )
            return None
        VerticalPortScan.logger.info(f"got {len(baselines)} partial baselines to merge from the baseline store")

        merged_baselines = pd.DataFrame()
        for partial_baseline in baselines:
            if partial_baseline.interval.end_time > self._beginning_of_today:
                # this case should only occur if we're running the previous days detection (usually occurs at 2am the next day)
                VerticalPortScan.logger.warn(
                    f"skipping partial baseline, end time is {partial_baseline.interval.end_time} while current run time is {self._now}"
                )
                continue
            baseline = pd.read_parquet(partial_baseline.data)
            merged_baselines = pd.concat([merged_baselines, baseline])

        # remove duplicate (dstVmId, timestamp) from merged_baselines
        merged_baselines = merged_baselines.drop_duplicates(subset=["dstVmId", "timestamp"], keep="last")

        if len(merged_baselines.index) == 0:
            return None

        combined_baseline = pd.DataFrame()

        # calculate duration
        now = pd.Timestamp(self._now, tz=timezone.utc)
        combined_baseline["duration"] = now - merged_baselines.groupby("dstVmId")["timestamp"].min()  # type: ignore (suppress linter warning)

        # list outlier datetimes
        combined_baseline["outlier_datetimes"] = (
            merged_baselines[
                merged_baselines.groupby("dstVmId")["cardinalityCount"]
                .transform(lambda x: np.abs(stats.zscore(x)))
                .gt(float(Config.OUTLIER_SIGMA), fill_value=0)
            ]
            .groupby("dstVmId", group_keys=False)["timestamp"]
            .apply(list)
        )

        # drop outliers and recalculate mean and std
        merged_baselines = merged_baselines[
            merged_baselines.groupby("dstVmId")["cardinalityCount"]
            .transform(lambda x: np.abs(stats.zscore(x)))
            .lt(float(Config.OUTLIER_SIGMA), fill_value=0)
        ]

        combined_baseline["mean"] = merged_baselines.groupby("dstVmId")["cardinalityCount"].mean()
        combined_baseline["std"] = merged_baselines.groupby("dstVmId")["cardinalityCount"].std()

        buffer = io.BytesIO()
        combined_baseline.to_parquet(buffer)
        VerticalPortScan.logger.info(
            f"combined partial baselines into a single baseline object with size {buffer.getbuffer().nbytes / 1_000}KB"
        )
        return buffer

    def _detect_events(
        self, user_config: UserConfigInfo, baseline: io.BytesIO, checkpoint: io.BytesIO | None
    ) -> Generator[EventInfo, Any, Any]:
        today_count = self._get_today_cardinality_count(user_config)
        if len(today_count.index) == 0:
            VerticalPortScan.logger.warn("got empty results for today's flow query, cannot detect events")
            return
        VerticalPortScan.logger.info(
            f"when detecting events, got {len(today_count.index)} cardinality counts for today"
        )
        VerticalPortScan.logger.debug(f"today query index: {today_count.index.values}")

        # drop computes in the exclude list
        excluded_computes = self._get_excluded_computes_from_config(user_config)
        if VerticalPortScan.logger.getEffectiveLevel() == logging.DEBUG:
            VerticalPortScan.logger.debug(f"got {len(excluded_computes)} computes from the exclusion list")
            VerticalPortScan.logger.debug(f"computes to exclude: {excluded_computes}")
        today_count = today_count[~today_count.index.isin(excluded_computes)]

        # drop any rows from the baseline where the duration < threshold
        computes_detecting_on = pd.read_parquet(baseline)
        if VerticalPortScan.logger.getEffectiveLevel() == logging.DEBUG:
            computes_dropped = computes_detecting_on[
                computes_detecting_on["duration"] < pd.Timedelta(days=float(Config.MIN_BASELINE_DURATION_DAYS))
            ]
            VerticalPortScan.logger.debug(
                f"dropping {len(computes_dropped.index)} computes due to min baseline duration"
            )
            VerticalPortScan.logger.debug(f"computes to drop: {computes_dropped.index.to_list()}")
        computes_detecting_on = computes_detecting_on[
            computes_detecting_on["duration"] >= pd.Timedelta(days=float(Config.MIN_BASELINE_DURATION_DAYS))
        ]
        VerticalPortScan.logger.info(f"when detecting events, got {len(computes_detecting_on.index)} valid baselines")
        VerticalPortScan.logger.debug(f"baseline index: {computes_detecting_on.index.values}")

        # inner join on the baseline, dropping any rows where the VM is not found in the baseline
        computes_detecting_on = computes_detecting_on.join(today_count, how="inner")
        VerticalPortScan.logger.info(f"detecting events on {len(computes_detecting_on.index)} computes")
        """computes_detecting_on should look like

                    today_cardinality_count     mean    std     duration    outlier_datetimes
        dstVmId
        vm1         24                          15      2       15 days     [date 1, date 2, ...]
        vm2         323                         500     45      30 days     NaN
        vm3         932                         210     300     30 days     [date 3, date 4, ...]
        ...
        """

        # adds "port_z_score" column
        computes_detecting_on = self._calculate_port_z_score(computes_detecting_on)
        VerticalPortScan.logger.info(f"calculated port z score for {len(computes_detecting_on.index)} computes")

        # drop any rows where the z score < threshold
        if VerticalPortScan.logger.getEffectiveLevel() == logging.DEBUG:
            computes_dropped = computes_detecting_on[
                computes_detecting_on["port_z_score"] < float(Config.MIN_PORT_Z_SCORE)
            ]
            VerticalPortScan.logger.debug(f"dropping {len(computes_dropped.index)} computes due to low z score")
            VerticalPortScan.logger.debug(f"computes to drop: {computes_dropped.index.to_list()}")
        computes_detecting_on = computes_detecting_on[
            computes_detecting_on["port_z_score"] >= float(Config.MIN_PORT_Z_SCORE)
        ]

        if len(computes_detecting_on.index) == 0:
            VerticalPortScan.logger.warn("cannot detect events, no computes remain after z score check")
            return

        # adds "unique_ports_used_today" column
        computes_detecting_on = self._calculate_unique_ports_today(computes_detecting_on, user_config.site_id)
        VerticalPortScan.logger.info(
            f"calculated unique ports used today for {len(computes_detecting_on.index)} computes"
        )

        # adds "tcp_icmp_error_percent" column
        computes_detecting_on = self._calculate_tcp_icmp_error(computes_detecting_on, user_config.site_id)
        VerticalPortScan.logger.info(f"calculated tcp / icmp errors for {len(computes_detecting_on.index)} computes")

        # adds "confidence" column, is a weighted combination of "port_z_score", "unique_ports_used_today", and "tcp_icmp_error_percent"
        computes_detecting_on = self._calculate_confidence(computes_detecting_on)
        VerticalPortScan.logger.info(f"calculated confidence for {len(computes_detecting_on.index)} computes")
        VerticalPortScan.logger.info(f"sample of computes dataframe:\n{computes_detecting_on.head()}")

        # if not reporting silent events, drop all rows with confidence < sensitivity
        VerticalPortScan.logger.info(f"{user_config.settings}")
        sensitivity_setting = user_config.settings.get_sensitivity_meter_setting()
        if sensitivity_setting is None:
            raise VerticalPortScanDetectionError("could not read sensitivity setting from user config store, got None")
        if Config.PERSIST_LOW_CONFIDENCE_EVENTS_AS_SILENT != "True":
            computes_detecting_on = computes_detecting_on[
                computes_detecting_on["confidence"] >= sensitivity_setting.sensitivity
            ]
            VerticalPortScan.logger.info(
                f"after dropping low confidence events, num events to report is {len(computes_detecting_on.index)}"
            )

        computes_detecting_on["is_silent"] = computes_detecting_on["confidence"] < sensitivity_setting.sensitivity

        # prepare a compute->ip lookup table for all events (if there are events)
        compute_to_ip_lookup = {}
        if len(computes_detecting_on.index) > 0:
            compute_to_ip_lookup = self._lookup_service.computes_to_ip_lookup(
                user_config.site_id, Interval(self._beginning_of_today, self._now)
            )
            VerticalPortScan.logger.info(f"preparing compute to ip lookup (size:{len(compute_to_ip_lookup)})")

        # create events to report
        for compute_id, row in computes_detecting_on.iterrows():
            VerticalPortScan.logger.info(f"creating event for compute: {compute_id}")
            yield self._create_event(str(compute_id), row, user_config.site_id, compute_to_ip_lookup)

    def _get_today_cardinality_count(self, user_config: UserConfigInfo) -> pd.DataFrame:
        midnight_to_now = Interval(self._beginning_of_today, self._now)
        flow_query_result = self._execute_port_scan_query(
            [midnight_to_now], user_config.site_id, Config.BASELINE_FLOW_QUERY_GRANULARITY
        )
        if len(flow_query_result.index) == 0:
            VerticalPortScan.logger.warn("got empty results for today's flow query, cannot detect events")
            return pd.DataFrame()

        # drop any duplicates (shouldn't happen due to the query always being < day and granularity being DAY, but worth checking for)
        flow_query_result = (
            flow_query_result.drop_duplicates(subset=["dstVmId"], keep="last")
            .drop(columns=["timestamp"])
            .set_index("dstVmId")
        )
        flow_query_result.columns = ["today_cardinality_count"]
        """flow_query_result should look like

                    today_cardinality_count
        dstVmId
        vm1         24
        vm2         323
        vm3         932
        ...
        """

        return flow_query_result

    def _calculate_port_z_score(self, computes_detecting_on: pd.DataFrame) -> pd.DataFrame:
        # calculate z score for each using a minimum std
        computes_detecting_on["port_z_score"] = (
            computes_detecting_on["today_cardinality_count"] - computes_detecting_on["mean"]
        ) / computes_detecting_on["std"].clip(float(Config.MIN_BASELINE_STD))

        return computes_detecting_on

    def _calculate_unique_ports_today(self, computes_detecting_on: pd.DataFrame, site_id: str) -> pd.DataFrame:
        vm_to_unique_ports = {}
        for compute_id, row in computes_detecting_on.iterrows():
            intervals_excluding_today = [
                Interval(
                    start_time=self._beginning_of_today - self._baseline_max_retention_time(),
                    end_time=self._beginning_of_today,
                )
            ]
            intervals_including_today = [
                Interval(start_time=self._beginning_of_today - self._baseline_max_retention_time(), end_time=self._now)
            ]

            if np.all(pd.notnull(row["outlier_datetimes"])) and len(row["outlier_datetimes"]) != 0:
                intervals_excluding_today = self._remove_outlier_datetimes_from_interval(
                    intervals_excluding_today[0], row["outlier_datetimes"]
                )
                intervals_including_today = self._remove_outlier_datetimes_from_interval(
                    intervals_including_today[0], row["outlier_datetimes"]
                )

            # TODO: refactor the below two queries to execute once per-run, rather than once per-compute
            #       similar refactor should be done in other detectors, such as traffic drop
            baseline_result = self._execute_port_scan_query(
                intervals_excluding_today, site_id, "all", [str(compute_id)]
            )

            total_result = self._execute_port_scan_query(intervals_including_today, site_id, "all", [str(compute_id)])

            if len(baseline_result.index) != 1:
                raise VerticalPortScanDetectionError(
                    f"when counting unique ports used, baseline_result result should only have one entry. It has: {len(baseline_result.index)}"
                )

            if len(total_result.index) != 1:
                raise VerticalPortScanDetectionError(
                    f"when counting unique ports used, total_result should only have one entry. It has: {len(total_result.index)}"
                )

            # can simply use iloc[0] here to grab the value, as we've asserted above that the size is 1 for both
            unique_ports_used_today = (total_result["cardinalityCount"] - baseline_result["cardinalityCount"]).iloc[0]

            if unique_ports_used_today > row["today_cardinality_count"]:
                unique_ports_used_today = row["today_cardinality_count"]

            vm_to_unique_ports[compute_id] = unique_ports_used_today

        computes_detecting_on["unique_ports_used_today"] = pd.Series(vm_to_unique_ports)

        return computes_detecting_on

    def _calculate_tcp_icmp_error(self, computes_detecting_on: pd.DataFrame, site_id: str) -> pd.DataFrame:
        tcp_icmp_errors_result = self._execute_tcp_icmp_error_query(
            Interval(start_time=self._beginning_of_today, end_time=self._now), site_id
        )

        # drop any duplicates (shouldn't happen due to the query always being < day and granularity being DAY, but worth checking for)
        if len(tcp_icmp_errors_result.index) != 0:
            tcp_icmp_errors_result = (
                tcp_icmp_errors_result.drop_duplicates(subset=["dstVmId"], keep="last")
                .drop(columns=["timestamp"])
                .set_index("dstVmId")
            )
            tcp_icmp_errors_result.columns = ["tcp_icmp_error_count"]
            computes_detecting_on["tcp_icmp_error_count"] = tcp_icmp_errors_result
        else:
            computes_detecting_on["tcp_icmp_error_count"] = 0

        computes_detecting_on["tcp_icmp_error_percent"] = (
            100 * computes_detecting_on["tcp_icmp_error_count"] / computes_detecting_on["today_cardinality_count"]
        )

        # replace any nan values (due to div by 0 error) with 0 and clip max % to 100
        computes_detecting_on["tcp_icmp_error_count"] = computes_detecting_on["tcp_icmp_error_count"].replace(np.nan, 0)
        computes_detecting_on["tcp_icmp_error_percent"] = (
            computes_detecting_on["tcp_icmp_error_percent"].replace(np.nan, 0).clip(upper=100)
        )

        return computes_detecting_on

    def _calculate_confidence(self, computes_detecting_on: pd.DataFrame) -> pd.DataFrame:
        # calculate confidence for each VM
        # weighted_quad(port_z_score) + weighted_quad(unique_ports_used_today) + weighted_quad(tcp_icmp_error_percent)
        MAX_INDIVIDUAL_SCORE = 100
        DEFAULT_SCORE = 50
        DEFAULT_STD = 3
        DEFAULT_UNIQUE_PORTS = 100
        DEFAULT_ERROR_PERCENTAGE = 25

        Z_SCORE_QUAD_COEF = DEFAULT_SCORE * DEFAULT_SCORE / DEFAULT_STD
        Z_SCORE_WEIGHT = 0.45
        UNIQUE_PORTS_QUAD_COEF = DEFAULT_SCORE * DEFAULT_SCORE / DEFAULT_UNIQUE_PORTS
        UNIQUE_PORTS_WEIGHT = 0.45
        ERROR_QUAD_COEF = DEFAULT_SCORE * DEFAULT_SCORE / DEFAULT_ERROR_PERCENTAGE
        ERROR_QUAD_WEIGHT = 0.10
        computes_detecting_on["confidence"] = (
            (
                computes_detecting_on["port_z_score"]
                .mul(Z_SCORE_QUAD_COEF)
                .pow(1.0 / 2)
                .clip(upper=MAX_INDIVIDUAL_SCORE)
                .mul(Z_SCORE_WEIGHT)
            )
            + (
                computes_detecting_on["unique_ports_used_today"]
                .mul(UNIQUE_PORTS_QUAD_COEF)
                .pow(1.0 / 2)
                .clip(upper=MAX_INDIVIDUAL_SCORE)
                .mul(UNIQUE_PORTS_WEIGHT)
            )
            + (
                computes_detecting_on["tcp_icmp_error_percent"]
                .mul(ERROR_QUAD_COEF)
                .pow(1.0 / 2)
                .clip(upper=MAX_INDIVIDUAL_SCORE)
                .mul(ERROR_QUAD_WEIGHT)
            )
        )

        return computes_detecting_on

    def _create_event(
        self, compute_id: str, compute_row: pd.Series, site_id: str, compute_to_ip_lookup: dict[str, str]
    ) -> EventInfo:
        src_computes_to_dst_port_result = self._execute_port_scan_source_query_with_compute_id(compute_id, site_id)
        if compute_id in compute_to_ip_lookup:
            result_with_ips = self._execute_port_scan_source_query_with_compute_ips(
                compute_to_ip_lookup[compute_id], site_id
            )
            if len(result_with_ips) > len(src_computes_to_dst_port_result):
                src_computes_to_dst_port_result = result_with_ips

        vm_viz_filter = VizFilter(
            key=VizFilterOption.VM_ID, values=[compute_id], criteria=Criteria.EQ, condition=Condition.OR
        )

        ports_viz_filter = VizFilter(
            key=VizFilterOption.FLOW_DSTPORT,
            values=src_computes_to_dst_port_result["dstPort"].drop_duplicates().to_list(),
            condition=Condition.OR,
            criteria=Criteria.EQ,
        )

        viz_filter = CompositeVizFilter([vm_viz_filter, ports_viz_filter], Condition.AND)

        src_computes_to_dst_ports = {}
        src_compute_to_timestamps = {}
        for _, row in src_computes_to_dst_port_result.iterrows():
            src_vm_id = row["srcVmId"]
            ports = src_computes_to_dst_ports.get(src_vm_id, [])
            ports.append(row["dstPort"])
            row_timestamp = pd.to_datetime(row["timestamp"], format="%Y-%m-%d %H:%M:%S", utc=True)
            timestamps = src_compute_to_timestamps.get(
                src_vm_id,
                (
                    pd.to_datetime(pd.Timestamp.min, utc=True),
                    pd.to_datetime(pd.Timestamp.max, utc=True),
                ),
            )
            timestamps = (max(row_timestamp, timestamps[0]), min(row_timestamp, timestamps[1]))
            src_computes_to_dst_ports[src_vm_id] = ports
            src_compute_to_timestamps[src_vm_id] = timestamps
        event_detail = self._create_event_detail(site_id, compute_id, src_computes_to_dst_ports, compute_row)

        # truncate src computes for scale reasons, the event message can get rejected by kafka for being too large
        # TODO: rather than truncating compute here, we should probably truncate src_computes_to_dst_port_result
        src_computes_ids = (
            src_computes_to_dst_port_result["srcVmId"]
            .drop_duplicates()
            .to_list()[: int(Config.EVENT_METADATA_COMPUTE_LIMIT)]
        )

        event_metadata = EventMetadata(
            kv_store={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_COMPUTES): ListString(src_computes_ids),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_COMPUTES): ListString([compute_id]),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_HOTSPOTS): ListString([compute_id]),
                EventKey.EVENT_METADATA_REPORT_KEY: JsonValue(json.dumps(event_detail.to_dict())),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): ListString(
                    [compute_to_ip_lookup[compute] for compute in src_computes_ids if compute in compute_to_ip_lookup]
                ),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): ListString(
                    [compute_to_ip_lookup[compute_id] if compute_id in compute_to_ip_lookup else ""]
                ),
                EventKey.EVENT_METADATA_VIZ_FILTER_KEY: JsonValue(viz_filter.to_json_str()),
            }
        )

        merge_plan = MergePlan(
            dedup_key="{}_VSCAN_{}".format("NSX_I", compute_id),
            metadata_key_merge_operation={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_COMPUTES): MergePolicy.SET_UNION_WITH_MAX,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): MergePolicy.KEEP_ORIGINAL,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): MergePolicy.KEEP_ORIGINAL,
            },
        )

        now_epoch_milli = int(self._now.timestamp() * 1000)

        affected_compute_ip = compute_to_ip_lookup[compute_id] if compute_id in compute_to_ip_lookup else None
        affected_compute = Compute(
            compute_type=ComputeType.COMPUTE_TYPE_VM,
            ip_address=affected_compute_ip,
            compute_id=compute_id,
            fqdn=None,
            site_id=site_id,
            ip_mask=IpMask.PRIVATE,
        )

        data_flows = []
        for src_compute, ports in src_computes_to_dst_ports.items():
            src_compute_ip = compute_to_ip_lookup.get(src_compute)
            start_time, end_time = src_compute_to_timestamps[src_compute]
            data_flows.append(
                DataFlow(
                    src_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address=src_compute_ip,
                        compute_id=src_compute,
                        fqdn=None,
                        site_id=site_id,
                        ip_mask=IpMask.PRIVATE,
                    ),
                    dst_compute=affected_compute,
                    transport_protocols=None,
                    application_protocols=None,
                    server_ports=[int(port) for port in ports],
                    start_time=int(start_time.timestamp() * 1000),
                    end_time=int(end_time.timestamp() * 1000),
                )
            )

        event_info = EventInfo(
            version="v1.1",
            detector_name=self.detector_name(),
            detection_start_time_epoch_milli=now_epoch_milli,
            detection_end_time_epoch_milli=now_epoch_milli,
            score=Score(round(compute_row["confidence"]), severity=20),
            event_metadata=event_metadata,
            merge_plan=merge_plan,
            mitre_tactic=Tactic.DISCOVERY,
            mitre_technique=Technique.NETWORK_SERVICE_SCANNING,
            silent=compute_row["is_silent"],
            site_ids=[site_id],
            org_id=Config.ORG_ID,
            affected_compute=affected_compute,
            is_attacker=False,
            is_client=False,
            data_flows=data_flows,
        )

        return event_info

    def _execute_port_scan_source_query_with_compute_id(self, compute_id: str, site_id: str) -> pd.DataFrame:
        filters = [
            Dimension("dstVmId") == compute_id,
            Dimension("algFlowType") != "DATA",
            Dimension("dstSiteId") == site_id,
            Dimension("isSrcInfoPresent") == True,  # noqa
        ]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity="day",
            intervals=[Interval(self._beginning_of_today, self._now)],
            aggregations={},
            dimensions=["srcVmId", "dstPort"],
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query)

    def _execute_port_scan_source_query_with_compute_ips(self, compute_ip: str, site_id: str) -> pd.DataFrame:
        filters = [
            Dimension("dstIps") == compute_ip,
            Dimension("algFlowType") != "DATA",
            Dimension("dstSiteId") == site_id,
            Dimension("isSrcInfoPresent") == True,  # noqa
        ]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity="day",
            intervals=[Interval(self._beginning_of_today, self._now)],
            aggregations={},
            dimensions=["srcVmId", "dstPort"],
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query)

    def _create_event_detail(
        self, site_id: str, compute_id: str, src_computes_to_dst_ports: dict[str, list[str]], compute_row: pd.Series
    ) -> PortScanVerticalEventDetail:
        interval_including_today = Interval(
            start_time=self._beginning_of_today - self._baseline_max_retention_time(), end_time=self._now
        )

        event_detail_query_result = self._execute_event_detail_query(interval_including_today, site_id, compute_id)
        event_detail_query_result["timestamp"] = pd.to_datetime(
            event_detail_query_result["timestamp"], format="%Y-%m-%d %H:%M:%S"
        )
        # convert to epoch milli
        event_detail_query_result["timestamp"] = event_detail_query_result["timestamp"].astype("int64") // 1_000_000

        port_count_series = Series(
            data=[int(elem) for elem in event_detail_query_result["cardinalityCount"].astype("int").to_list()],
            labels=event_detail_query_result["timestamp"].astype("str").to_list(),
            mean=float(compute_row["mean"]),
            stddev=float(compute_row["std"]),
        )

        scanned_vm_details = []
        total_ports = 0
        for src_compute, dst_ports in src_computes_to_dst_ports.items():
            # TODO: might want to add a limit to the number of ports per-scanned vm as well as the number of scanned
            #       VMs we report, such that kafka topic does not reject the event
            scanned_vm_detail = ScannedVmDetails(
                vm_id=src_compute, ports=[int(port) for port in dst_ports], port_protocols=None
            )
            scanned_vm_details.append(scanned_vm_detail)
            total_ports += len(dst_ports)
        unique_ports_percent = min(100 * compute_row["unique_ports_used_today"] / total_ports, 100)

        return PortScanVerticalEventDetail(
            port_count=port_count_series,
            scanner_info=scanned_vm_details,
            total_ports=total_ports,
            total_unique_ports=int(compute_row["unique_ports_used_today"]),
            unique_ports_percent=int(unique_ports_percent),
            report_type=self.detector_name().name,
            viz_filter_elements=None,
        )

    def _get_excluded_computes_from_config(self, config: UserConfigInfo) -> list[str]:
        if config.exclusion_settings == {}:
            return []

        excluded_computes = []
        for setting in config.exclusion_settings.exclusions:
            if setting.label == ExclusionLabel.DESTINATION:
                # setting.excluded_items_info.value can technically be any type, but here we're assuming it's a list of str
                excluded_entity_ids: list[str] = setting.excluded_items_info.value  # type: ignore (suppress linter warning)
                if setting.type == ExclusionTypeInfo.VM:
                    excluded_computes += excluded_entity_ids
                elif setting.type == ExclusionTypeInfo.GROUP:
                    excluded_computes += self._lookup_service.groups_to_computes_lookup(
                        config.site_id, excluded_entity_ids
                    )
        return excluded_computes

    def _remove_outlier_datetimes_from_interval(
        self, interval: Interval, outlier_datetimes: list[str]
    ) -> list[Interval]:
        updated_intervals = []
        current_start = interval.start_time
        parsed_datetimes = pd.to_datetime(outlier_datetimes, format="%Y-%m-%d %H:%M:%S")
        for outlier in parsed_datetimes.sort_values():
            if outlier < interval.start_time:
                continue
            if outlier >= interval.end_time:
                break
            if outlier != current_start:
                updated_intervals.append(Interval(start_time=current_start, end_time=outlier))
            current_start = outlier + timedelta(days=1)
        if current_start != interval.end_time:
            updated_intervals.append(Interval(start_time=current_start, end_time=interval.end_time))

        return updated_intervals

    def _execute_tcp_icmp_error_query(self, interval: Interval, site_id: str) -> pd.DataFrame:
        error_filters = [
            Bound("dstTcpIncomplete3WH", "0", lowerStrict=True),
            Bound("srcIcmpDestUnreachable", "0", lowerStrict=True),
            Bound("dstIcmpDestUnreachable", "0", lowerStrict=True),
        ]

        filters = [
            Dimension("algFlowType") != "DATA",
            Dimension("dstComputeType") == "VM",
            Dimension("dstSiteId") == site_id,
            Dimension("isDstInfoPresent") == True,  # noqa
            Filter(type="or", fields=error_filters),
        ]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity="day",
            intervals=[interval],
            aggregations={"cardinalityCount": cardinality("dstPort")},
            dimensions=["dstVmId"],
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query)

    def _execute_event_detail_query(self, interval: Interval, site_id: str, compute_id: str) -> pd.DataFrame:
        filters = [Dimension("dstVmId") == compute_id, Dimension("dstSiteId") == site_id]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity="day",
            intervals=[interval],
            aggregations={"cardinalityCount": cardinality("dstPort")},
            dimensions=["dstVmId"],
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query)

    def _get_detector_run_status(
        self, status_label: DetectorStatusLabel, site_id: str, last_successful_run_time: int, message: str = ""
    ) -> DetectorRunStatus:
        VerticalPortScan.logger.debug(f"Setting detector run status for {self.detector_name().name}")
        max_baseline_duration_days = self._baseline_service.get_max_baseline_duration_days(
            self.detector_name().name, Config.ORG_ID, site_id
        )
        remaining_baseline_duration_days = max(int(Config.MIN_BASELINE_DURATION_DAYS) - max_baseline_duration_days, 0)
        status_label = (
            DetectorStatusLabel.NOT_ENOUGH_BASELINE
            if (status_label == DetectorStatusLabel.SUCCESS and remaining_baseline_duration_days > 0)
            else status_label
        )

        return DetectorRunStatus(
            status_label=status_label,
            last_successful_run_time=last_successful_run_time,
            max_baseline_duration_days=max_baseline_duration_days,
            remaining_baseline_duration_days=remaining_baseline_duration_days,
            message=message,
        )
