import io
import unittest
import unittest.mock
from datetime import date, datetime, timedelta, timezone
from unittest.mock import ANY, Mock

import numpy as np
import pandas as pd
from confluent_kafka import Producer

from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import Interval, PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.common_model import DetectorType
from common.utils.event.core import EventService
from common.utils.event.data import (
    Compute,
    ComputeType,
    DataFlow,
    IpMask,
    MergePolicy,
    Tactic,
    Technique,
)
from common.utils.event.data import Data as EventDataService
from common.utils.flow.core import FlowService
from common.utils.flow.data import FlowGroupByQuery
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    AnomalyExclusionSetting,
    AnomalyExclusionSettings,
    DetectorRunStatus,
    DetectorSettings,
    DetectorStatusLabel,
    ExcludedItemsInfo,
    ExclusionLabel,
    ExclusionTypeInfo,
    UserConfigInfo,
)
from verticalportscan.detector.config import Config
from verticalportscan.detector.core import VerticalPortScan
from vmware.nsx.intelligence.nta.v1 import nta_pb2


class TestCore(unittest.TestCase):
    def setUp(self):
        self._mock_baseline_service = Mock(BaselineService)
        self._mock_checkpoint_service = Mock(CheckpointService)
        self._mock_user_config_service = Mock(UserConfigService)
        self._mock_event_service = Mock(EventService)
        self._mock_flow_service = Mock(FlowService)
        self._mock_lookup_service = Mock(LookupService)
        self.detector = VerticalPortScan(
            self._mock_baseline_service,
            self._mock_checkpoint_service,
            self._mock_user_config_service,
            self._mock_event_service,
            self._mock_flow_service,
            self._mock_lookup_service,
            datetime.now(),
        )

        self._event_data_service = EventDataService(Producer({}))

    def test_detector_name(self):
        self.assertEqual(DetectorType.VERTICAL_PORT_SCAN, self.detector.detector_name())

    def test_adjust_missing_intervals(self):
        today = datetime.combine(date.today(), datetime.min.time())
        intervals_1 = [
            Interval(today - timedelta(days=1), today),
            Interval(today - timedelta(days=2), today - timedelta(days=1)),
            Interval(today - timedelta(days=3), today - timedelta(days=2)),
            Interval(today - timedelta(days=4), today - timedelta(days=3)),
            Interval(today - timedelta(days=5), today - timedelta(days=4)),
            Interval(today - timedelta(days=6), today - timedelta(days=5)),
            Interval(today - timedelta(days=7), today - timedelta(days=6)),
            Interval(today - timedelta(days=8), today - timedelta(days=7)),
            Interval(today - timedelta(days=9), today - timedelta(days=8)),
            Interval(today - timedelta(days=10), today - timedelta(days=9)),
            Interval(today - timedelta(days=11), today - timedelta(days=10)),
        ]
        expected_intervals_1 = [
            Interval(today - timedelta(days=1), today),
            Interval(today - timedelta(days=2), today - timedelta(days=1)),
            Interval(today - timedelta(days=3), today - timedelta(days=2)),
            Interval(today - timedelta(days=4), today - timedelta(days=3)),
            Interval(today - timedelta(days=5), today - timedelta(days=4)),
            Interval(today - timedelta(days=6), today - timedelta(days=5)),
            Interval(today - timedelta(days=7), today - timedelta(days=6)),
        ]

        intervals_2 = [
            Interval(today - timedelta(days=5), today - timedelta(days=4)),
            Interval(today - timedelta(days=6), today - timedelta(days=5)),
            Interval(today - timedelta(days=10), today - timedelta(days=6)),
        ]
        expected_intervals_2 = [
            Interval(today - timedelta(days=5), today - timedelta(days=4)),
            Interval(today - timedelta(days=6), today - timedelta(days=5)),
            Interval(today - timedelta(days=7), today - timedelta(days=6)),
        ]

        intervals_3 = [Interval(today - timedelta(days=30), today)]
        expected_intervals_3 = [
            Interval(today - timedelta(days=1), today),
            Interval(today - timedelta(days=2), today - timedelta(days=1)),
            Interval(today - timedelta(days=3), today - timedelta(days=2)),
            Interval(today - timedelta(days=4), today - timedelta(days=3)),
            Interval(today - timedelta(days=5), today - timedelta(days=4)),
            Interval(today - timedelta(days=6), today - timedelta(days=5)),
            Interval(today - timedelta(days=7), today - timedelta(days=6)),
        ]

        intervals_4 = [
            Interval(today - timedelta(days=8), today - timedelta(days=7)),
            Interval(today - timedelta(days=9), today - timedelta(days=8)),
            Interval(today - timedelta(days=10), today - timedelta(days=9)),
            Interval(today - timedelta(days=11), today - timedelta(days=10)),
        ]
        expected_intervals_4 = []

        adjusted_intervals_1 = self.detector._get_baseline_intervals(intervals_1)
        self.assertEqual(expected_intervals_1, adjusted_intervals_1)

        adjusted_intervals_2 = self.detector._get_baseline_intervals(intervals_2)
        self.assertEqual(expected_intervals_2, adjusted_intervals_2)

        adjusted_intervals_3 = self.detector._get_baseline_intervals(intervals_3)
        self.assertEqual(expected_intervals_3, adjusted_intervals_3)

        adjusted_intervals_4 = self.detector._get_baseline_intervals(intervals_4)
        self.assertEqual(expected_intervals_4, adjusted_intervals_4)

        self.assertEqual([], self.detector._get_baseline_intervals([]))

    def test_get_query_intervals_from_missing_intervals(self):
        today = datetime.now()
        beginning_of_today = datetime.combine(date.today(), datetime.min.time())
        intervals_1 = [
            Interval(today - timedelta(days=10), beginning_of_today),
            Interval(today - timedelta(days=13), today - timedelta(days=11)),
            Interval(today - timedelta(days=21), today - timedelta(days=19)),
        ]
        expected_intervals_1 = [
            Interval(beginning_of_today - timedelta(days=1), beginning_of_today),
            Interval(beginning_of_today - timedelta(days=2), beginning_of_today - timedelta(days=1)),
            Interval(beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=2)),
            Interval(beginning_of_today - timedelta(days=4), beginning_of_today - timedelta(days=3)),
            Interval(beginning_of_today - timedelta(days=5), beginning_of_today - timedelta(days=4)),
            Interval(beginning_of_today - timedelta(days=6), beginning_of_today - timedelta(days=5)),
            Interval(beginning_of_today - timedelta(days=7), beginning_of_today - timedelta(days=6)),
        ]

        query_interval_1 = self.detector._get_baseline_intervals(intervals_1)
        self.assertEqual(expected_intervals_1, query_interval_1)

    def test_combine_baseline(self):
        baselines = [
            [
                ["vm1", 0, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vm2", 5, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vm3", 10, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vm4", 1, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vm5", 15, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ],
            [
                ["vm1", 0, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vm2", 5, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vm3", 20, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vm4", 6, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vm5", 30, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
            ],
            [
                ["vm1", 0, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vm2", 5, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vm3", 30, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vm4", 400, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vm5", 40, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
            ],
            [
                ["vm1", 0, datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],
                ["vm2", 5, datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],
                ["vm3", 20, datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],
                ["vm4", 3, datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],
                ["vm5", 20, datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],
            ],
            [
                ["vm1", 0, datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
                ["vm2", 5, datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
                ["vm3", 20, datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
                ["vm4", 2, datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
                ["vm5", 5, datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
            ],
        ]
        baselines = [
            pd.DataFrame(baseline, columns=["dstVmId", "cardinalityCount", "timestamp"]) for baseline in baselines
        ]

        partial_baselines = []
        for baseline in baselines:
            buffer = io.BytesIO()
            baseline["timestamp"] = pd.to_datetime(baseline["timestamp"], format="%Y-%m-%d %H:%M:%S")
            baseline.to_parquet(buffer)
            partial_baselines.append(
                PartialBaseline(
                    Interval(datetime.now() - timedelta(days=5), datetime.now() - timedelta(days=4)), buffer
                )
            )

        result = self.detector._combine_baseline_objects(partial_baselines)
        self.assertIsNotNone(result)

        baseline = pd.read_parquet(result)  # type: ignore
        self.assertEqual(0, baseline.loc["vm1"]["mean"])
        self.assertEqual(5, baseline.loc["vm2"]["mean"])
        self.assertEqual(20, baseline.loc["vm3"]["mean"])

        self.assertEqual(0, baseline.loc["vm1"]["std"])
        self.assertEqual(0, baseline.loc["vm2"]["std"])
        self.assertAlmostEqual(7.0711, baseline.loc["vm3"]["std"], places=4)

    def test_get_today_cardinality_count(self):
        result = [
            ["vm1", 0, datetime(year=2023, month=1, day=1)],
            ["vm2", 5, datetime(year=2023, month=1, day=1)],
            ["vm3", 10, datetime(year=2023, month=1, day=1)],
            ["vm4", 1, datetime(year=2023, month=1, day=1)],
            ["vm5", 15, datetime(year=2023, month=1, day=1)],
            ["vm10", 0, datetime(year=2023, month=1, day=2)],
            ["vm11", 5, datetime(year=2023, month=1, day=2)],
            ["vm12", 20, datetime(year=2023, month=1, day=2)],
            ["vm13", 6, datetime(year=2023, month=1, day=2)],
            ["vm14", 30, datetime(year=2023, month=1, day=2)],
            ["vm6", 0, datetime(year=2023, month=1, day=3)],
            ["vm7", 5, datetime(year=2023, month=1, day=3)],
            ["vm8", 30, datetime(year=2023, month=1, day=3)],
            ["vm9", 400, datetime(year=2023, month=1, day=3)],
            ["vm9", 40, datetime(year=2023, month=1, day=4)],
        ]

        result_df = pd.DataFrame(result, columns=["dstVmId", "cardinalityCount", "timestamp"])
        self._mock_flow_service.group_by = lambda _query: result_df

        exclusion_vm = AnomalyExclusionSetting(
            label=ExclusionLabel.DESTINATION,
            type=ExclusionTypeInfo.VM,
            excluded_items_info=ExcludedItemsInfo(["vm10", "vm11", "vm14"]),
        )

        exclusion_group = AnomalyExclusionSetting(
            label=ExclusionLabel.DESTINATION,
            type=ExclusionTypeInfo.GROUP,
            excluded_items_info=ExcludedItemsInfo(["group1", "group2"]),
        )

        self._mock_lookup_service.groups_to_computes_lookup = lambda _site_id, _group_ids: ["vm13", "vm6"]

        user_config = UserConfigInfo(
            site_id="site1",
            name="VSCAN",
            org_id="org_1",
            version="1",
            last_modified_time=1234,
            enabled=True,
            silent=False,
            source="source",
            block=False,
            block_confidence_threshold=1234,
            mitre_tactics=[],
            mitre_techniques=[],
            settings=DetectorSettings([]),
            exclusion_settings=AnomalyExclusionSettings([exclusion_vm, exclusion_group]),
            run_status=DetectorRunStatus(DetectorStatusLabel.NOT_STARTED),
        )

        today_count = self.detector._get_today_cardinality_count(user_config)

        self.assertEqual(14, len(today_count.index))
        self.assertListEqual(
            today_count.index.to_list(),
            ["vm1", "vm2", "vm3", "vm4", "vm5", "vm10", "vm11", "vm12", "vm13", "vm14", "vm6", "vm7", "vm8", "vm9"],
        )

    def test_calculate_port_z_score(self):
        df = [
            [1423, timedelta(days=30), np.nan, 1000, 324],
            [63, timedelta(days=23), [], 5, 1],
            [43, timedelta(days=25), [], 40, 20],
            [0, timedelta(days=30), [], 500, 30],
            [1.3, timedelta(days=30), [], 0.5, 1.2],
        ]
        index = ["vm1", "vm2", "vm3", "vm4", "vm5"]

        input_df = pd.DataFrame(
            df, index, columns=["today_cardinality_count", "duration", "outlier_datetimes", "mean", "std"]
        )

        result_df = self.detector._calculate_port_z_score(input_df)
        self.assertAlmostEqual(1.306, result_df.loc["vm1"]["port_z_score"], places=3)
        self.assertEqual(58, result_df.loc["vm2"]["port_z_score"])
        self.assertEqual(0.15, result_df.loc["vm3"]["port_z_score"])
        self.assertAlmostEqual(-16.67, result_df.loc["vm4"]["port_z_score"], places=2)
        self.assertAlmostEqual(0.6667, result_df.loc["vm5"]["port_z_score"], places=4)

    def test_remove_outlier_datetimes_from_interval(self):
        intervals = [
            Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=30)),
            Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=30)),
            Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=30)),
            Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=30)),
            Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=30)),
        ]

        outlier_datetimes = [
            [
                datetime(year=2023, month=1, day=10).isoformat(),
                datetime(year=2023, month=1, day=11).isoformat(),
                datetime(year=2023, month=1, day=13).isoformat(),
                datetime(year=2023, month=1, day=15).isoformat(),
            ],
            [(datetime(year=2023, month=1, day=1) + timedelta(days=30)).isoformat()],
            [],
            [(datetime(year=2023, month=1, day=1) + timedelta(days=29)).isoformat()],
            [datetime(year=2023, month=1, day=1).isoformat()],
            [datetime(year=2022, month=1, day=1).isoformat()],
        ]

        results = [
            self.detector._remove_outlier_datetimes_from_interval(intervals[i], outlier_datetimes[i])
            for i in range(len(intervals))
        ]

        expected_results = [
            [
                Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=10)),
                Interval(datetime(year=2023, month=1, day=12), datetime(year=2023, month=1, day=13)),
                Interval(datetime(year=2023, month=1, day=14), datetime(year=2023, month=1, day=15)),
                Interval(datetime(year=2023, month=1, day=16), datetime(year=2023, month=1, day=31)),
            ],
            [Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=30))],
            [Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=30))],
            [Interval(datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1) + timedelta(days=29))],
            [Interval(datetime(year=2023, month=1, day=2), datetime(year=2023, month=1, day=1) + timedelta(days=30))],
        ]

        for i in range(len(intervals)):
            self.assertListEqual(expected_results[i], results[i])

    def test_calculate_unique_ports_today(self):
        beginning_of_today = datetime.combine(date.today(), datetime.min.time())
        df = [
            [35, timedelta(days=30), np.nan, 14, 2],
            [10, timedelta(days=30), np.nan, 14, 2],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
        ]
        index = ["vm1", "vm2", "vm3"]
        columns = ["today_cardinality_count", "duration", "outlier_datetimes", "mean", "std"]

        computes_detecting_on = pd.DataFrame(df, index, columns=columns)

        self._mock_flow_service.group_by = mock_unique_ports_flow_query

        result = self.detector._calculate_unique_ports_today(computes_detecting_on, "default")

        self.assertEqual(20, result.loc["vm1"]["unique_ports_used_today"])
        self.assertEqual(10, result.loc["vm2"]["unique_ports_used_today"])
        self.assertEqual(0, result.loc["vm3"]["unique_ports_used_today"])

    def test_calculate_tcp_icmp_error(self):
        result = [
            ["vm1", 0, datetime(year=2023, month=1, day=1)],
            ["vm2", 5, datetime(year=2023, month=1, day=1)],
            ["vm3", 10, datetime(year=2023, month=1, day=1)],
            ["vm4", 1, datetime(year=2023, month=1, day=1)],
            ["vm5", 15, datetime(year=2023, month=1, day=1)],
            ["vm10", 0, datetime(year=2023, month=1, day=2)],
            ["vm6", 0, datetime(year=2023, month=1, day=3)],
            ["vm7", 5, datetime(year=2023, month=1, day=3)],
            ["vm8", 30, datetime(year=2023, month=1, day=3)],
            ["vm9", 400, datetime(year=2023, month=1, day=3)],
            ["vm9", 40, datetime(year=2023, month=1, day=4)],
        ]

        result_df = pd.DataFrame(result, columns=["dstVmId", "cardinalityCount", "timestamp"])
        self._mock_flow_service.group_by = lambda _query: result_df

        beginning_of_today = datetime.combine(date.today(), datetime.min.time())
        df = [
            [35, timedelta(days=30), np.nan, 14, 2],
            [10, timedelta(days=30), np.nan, 14, 2],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
            [
                35,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
            [
                0,
                timedelta(days=30),
                [beginning_of_today - timedelta(days=3), beginning_of_today - timedelta(days=6)],
                14,
                2,
            ],
        ]
        index = ["vm1", "vm2", "vm3", "vm4", "vm5", "vm6", "vm7", "vm8", "vm9", "vm10"]
        columns = ["today_cardinality_count", "duration", "outlier_datetimes", "mean", "std"]

        computes_detecting_on = pd.DataFrame(df, index, columns=columns)

        result = self.detector._calculate_tcp_icmp_error(computes_detecting_on, "default")

        self.assertEqual(0, result.loc["vm1"]["tcp_icmp_error_count"])
        self.assertEqual(15, result.loc["vm5"]["tcp_icmp_error_count"])
        self.assertEqual(40, result.loc["vm9"]["tcp_icmp_error_count"])

        self.assertEqual(0, result.loc["vm1"]["tcp_icmp_error_percent"])
        self.assertEqual(50, result.loc["vm2"]["tcp_icmp_error_percent"])
        self.assertAlmostEqual(42.857, result.loc["vm5"]["tcp_icmp_error_percent"], places=3)
        self.assertEqual(100, result.loc["vm9"]["tcp_icmp_error_percent"])
        self.assertEqual(0, result.loc["vm10"]["tcp_icmp_error_percent"])

        result = [
            ["vm1", 0, datetime(year=2023, month=1, day=1)],
            ["vm2", 5, datetime(year=2023, month=1, day=1)],
            ["vm3", 10, datetime(year=2023, month=1, day=1)],
            ["vm4", 1, datetime(year=2023, month=1, day=1)],
        ]

        result_df = pd.DataFrame(result, columns=["dstVmId", "cardinalityCount", "timestamp"])
        self._mock_flow_service.group_by = lambda _query: result_df

        result = self.detector._calculate_tcp_icmp_error(computes_detecting_on, "default")

        self.assertEqual(0, result.loc["vm5"]["tcp_icmp_error_count"])
        self.assertEqual(0, result.loc["vm5"]["tcp_icmp_error_percent"])
        self.assertEqual(0, result.loc["vm9"]["tcp_icmp_error_count"])
        self.assertEqual(0, result.loc["vm9"]["tcp_icmp_error_percent"])

    def test_calculate_confidence(self):
        df = [
            [35, timedelta(days=30), np.nan, 14, 2, 0, 0, 0],
            [10, timedelta(days=30), np.nan, 14, 2, 10, 0, 0],
            [10, timedelta(days=30), np.nan, 14, 2, 0, 10, 0],
            [10, timedelta(days=30), np.nan, 14, 2, 0, 0, 10],
            [10, timedelta(days=30), np.nan, 14, 2, 15, 20, 40],
            [10, timedelta(days=30), np.nan, 14, 2, 1.5, 2.3, 4.2],
            [10, timedelta(days=30), np.nan, 14, 2, 3, 100, 25],
            [10, timedelta(days=30), np.nan, 14, 2, 3, 0, 0],
            [10, timedelta(days=30), np.nan, 14, 2, 0, 100, 0],
            [10, timedelta(days=30), np.nan, 14, 2, 0, 0, 25],
        ]
        index = ["vm1", "vm2", "vm3", "vm4", "vm5", "vm6", "vm7", "vm8", "vm9", "vm10"]
        columns = [
            "today_cardinality_count",
            "duration",
            "outlier_datetimes",
            "mean",
            "std",
            "port_z_score",
            "unique_ports_used_today",
            "tcp_icmp_error_percent",
        ]

        computes_detecting_on = pd.DataFrame(df, index, columns=columns)

        result = self.detector._calculate_confidence(computes_detecting_on)

        self.assertEqual(0, result.loc["vm1"]["confidence"])
        self.assertAlmostEqual(61.387, result.loc["vm5"]["confidence"], places=3)
        self.assertAlmostEqual(21.372, result.loc["vm6"]["confidence"], places=3)
        self.assertEqual(50, result.loc["vm7"]["confidence"])
        self.assertEqual(22.5, result.loc["vm8"]["confidence"])
        self.assertEqual(22.5, result.loc["vm9"]["confidence"])
        self.assertEqual(5, result.loc["vm10"]["confidence"])

    def test_create_event(self):
        self._mock_flow_service.group_by = mock_create_event_query

        compute_to_ip = {
            "vm1": "192.168.0.1",
            "vm2": "192.168.0.2",
            "vm3": "192.168.0.3",
            "vm4": "192.168.0.4",
            "vm5": "192.168.0.5",
            "vm100": "192.168.0.100",
            "vm200": "192.168.0.200",
        }

        self._mock_lookup_service.computes_to_ip_lookup = lambda _site_id, _interval: compute_to_ip

        df = [
            [35, True, 124, 23, 130],
            [10, True, 423, 12, 400],
            [100, False, 30, 14, 28],
            [85, False, 32, 7, 33],
            [90, False, 3.2, 2, 5],
            [10, True, 44, 6, 28],
            [20, True, 93, 24, 99],
            [25, True, 52, 5, 84],
            [35, True, 4, 5, 10],
            [0, True, 3, 1, 1],
        ]
        index = ["vm1", "vm2", "vm3", "vm4", "vm5", "vm6", "vm7", "vm8", "vm9", "vm10"]
        columns = ["confidence", "is_silent", "mean", "std", "unique_ports_used_today"]

        computes_detecting_on = pd.DataFrame(df, index, columns=columns)

        for compute_id, row in computes_detecting_on.iterrows():
            result = self.detector._create_event(
                str(compute_id), row, "default", self._mock_lookup_service.computes_to_ip_lookup(None, None)
            )

            self.assertEqual(self.detector.detector_name(), result.detector_name)

            self.assertEqual(row["confidence"], result.score.confidence)
            self.assertEqual(20, result.score.severity)
            self.assertEqual(
                ["vm100", "vm200", "vm300", "vm400"],
                result.event_metadata.kv_store.get("METADATA_KEY_SRC_COMPUTES").list_string_value,  # type: ignore
            )
            self.assertEqual(
                [compute_id],
                result.event_metadata.kv_store.get("METADATA_KEY_DST_COMPUTES").list_string_value,  # type: ignore
            )
            self.assertEqual(
                [compute_id],
                result.event_metadata.kv_store.get("METADATA_KEY_HOTSPOTS").list_string_value,  # type: ignore
            )
            self.assertEqual(
                ["192.168.0.100", "192.168.0.200"],
                result.event_metadata.kv_store.get("METADATA_KEY_SRC_IPS").list_string_value,  # type: ignore
            )
            self.assertEqual(
                [compute_to_ip[str(compute_id)] if str(compute_id) in compute_to_ip else ""],
                result.event_metadata.kv_store.get("METADATA_KEY_DST_IPS").list_string_value,  # type: ignore
            )

            self.assertEqual("NSX_I_VSCAN_{}".format(compute_id), result.merge_plan.dedup_key)
            self.assertEqual(
                MergePolicy.SET_UNION_WITH_MAX,
                result.merge_plan.metadata_key_merge_operation.get("METADATA_KEY_SRC_COMPUTES"),
            )
            self.assertEqual(
                MergePolicy.KEEP_ORIGINAL, result.merge_plan.metadata_key_merge_operation.get("METADATA_KEY_SRC_IPS")
            )
            self.assertEqual(
                MergePolicy.KEEP_ORIGINAL, result.merge_plan.metadata_key_merge_operation.get("METADATA_KEY_DST_IPS")
            )

            self.assertEqual(Tactic.DISCOVERY, result.mitre_tactic)
            self.assertEqual(Technique.NETWORK_SERVICE_SCANNING, result.mitre_technique)
            self.assertEqual(row["is_silent"], result.silent)
            self.assertEqual(["default"], result.site_ids)

            expected_data_flows = [
                DataFlow(
                    src_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address="192.168.0.100",
                        compute_id="vm100",
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    dst_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address=compute_to_ip.get(str(compute_id)),
                        compute_id=str(compute_id),
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    transport_protocols=None,
                    application_protocols=None,
                    server_ports=[22],
                    start_time=ANY,
                    end_time=ANY,
                ),
                DataFlow(
                    src_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address="192.168.0.200",
                        compute_id="vm200",
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    dst_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address=compute_to_ip.get(str(compute_id)),
                        compute_id=str(compute_id),
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    transport_protocols=None,
                    application_protocols=None,
                    server_ports=[80, 22, 9000, 443],
                    start_time=ANY,
                    end_time=ANY,
                ),
                DataFlow(
                    src_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address=None,
                        compute_id="vm300",
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    dst_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address=compute_to_ip.get(str(compute_id)),
                        compute_id=str(compute_id),
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    transport_protocols=None,
                    application_protocols=None,
                    server_ports=[22],
                    start_time=ANY,
                    end_time=ANY,
                ),
                DataFlow(
                    src_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address=None,
                        compute_id="vm400",
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    dst_compute=Compute(
                        compute_type=ComputeType.COMPUTE_TYPE_VM,
                        ip_address=compute_to_ip.get(str(compute_id)),
                        compute_id=str(compute_id),
                        fqdn=None,
                        site_id="default",
                        ip_mask=IpMask.PRIVATE,
                    ),
                    transport_protocols=None,
                    application_protocols=None,
                    server_ports=[22],
                    start_time=ANY,
                    end_time=ANY,
                ),
            ]
            self.assertEqual(result.data_flows, expected_data_flows)

            pb_msg = self._event_data_service._info_to_proto(result)

            self.assertEqual(row["confidence"], pb_msg.score.confidence)
            self.assertEqual(20, pb_msg.score.severity)
            self.assertEqual(
                ["vm100", "vm200", "vm300", "vm400"],
                pb_msg.event_metadata.kv_store.get("METADATA_KEY_SRC_COMPUTES").list_string_value.list_string_value,  # type: ignore
            )
            self.assertEqual(
                [compute_id],
                pb_msg.event_metadata.kv_store.get("METADATA_KEY_DST_COMPUTES").list_string_value.list_string_value,  # type: ignore
            )
            self.assertEqual(
                [compute_id],
                pb_msg.event_metadata.kv_store.get("METADATA_KEY_HOTSPOTS").list_string_value.list_string_value,  # type: ignore
            )
            self.assertEqual(
                ["192.168.0.100", "192.168.0.200"],
                pb_msg.event_metadata.kv_store.get("METADATA_KEY_SRC_IPS").list_string_value.list_string_value,  # type: ignore
            )
            self.assertEqual(
                [compute_to_ip[str(compute_id)] if str(compute_id) in compute_to_ip else ""],
                pb_msg.event_metadata.kv_store.get("METADATA_KEY_DST_IPS").list_string_value.list_string_value,  # type: ignore
            )

            self.assertEqual("NSX_I_VSCAN_{}".format(compute_id), pb_msg.merge_plan.dedup_key)
            self.assertEqual(
                nta_pb2.MergePolicy.MERGE_POLICY_SET_UNION_WITH_MAX,
                pb_msg.merge_plan.metadata_key_merge_operation.get("METADATA_KEY_SRC_COMPUTES"),
            )
            self.assertEqual(
                nta_pb2.MergePolicy.MERGE_POLICY_KEEP_ORIGINAL,
                pb_msg.merge_plan.metadata_key_merge_operation.get("METADATA_KEY_SRC_IPS"),
            )
            self.assertEqual(
                nta_pb2.MergePolicy.MERGE_POLICY_KEEP_ORIGINAL,
                pb_msg.merge_plan.metadata_key_merge_operation.get("METADATA_KEY_DST_IPS"),
            )

            self.assertEqual(Tactic.DISCOVERY.name, pb_msg.mitre_tactic)
            self.assertEqual(Technique.NETWORK_SERVICE_SCANNING.name, pb_msg.mitre_technique)
            self.assertEqual(row["is_silent"], pb_msg.silent)
            self.assertEqual(["default"], pb_msg.site_ids)

        # try again, with limit applied
        original_limit = Config.EVENT_METADATA_COMPUTE_LIMIT
        Config.EVENT_METADATA_COMPUTE_LIMIT = "2"
        for compute_id, row in computes_detecting_on.iterrows():
            result = self.detector._create_event(
                str(compute_id), row, "default", self._mock_lookup_service.computes_to_ip_lookup(None, None)
            )
            self.assertEqual(
                ["vm100", "vm200"],
                result.event_metadata.kv_store.get("METADATA_KEY_SRC_COMPUTES").list_string_value,  # type: ignore
            )
            self.assertEqual(
                ["192.168.0.100", "192.168.0.200"],
                result.event_metadata.kv_store.get("METADATA_KEY_SRC_IPS").list_string_value,  # type: ignore
            )
        # reset limit for other tests
        Config.EVENT_METADATA_COMPUTE_LIMIT = original_limit

    def test_generate_baselines(self):
        result = [
            ["vm1", 0, datetime(year=2023, month=1, day=1)],
            ["vm2", 5, datetime(year=2023, month=1, day=1)],
            ["vm3", 10, datetime(year=2023, month=1, day=1)],
            ["vm4", 1, datetime(year=2023, month=1, day=1)],
            ["vm5", 15, datetime(year=2023, month=1, day=1)],
            ["vm10", 0, datetime(year=2023, month=1, day=2)],
            ["vm11", 5, datetime(year=2023, month=1, day=2)],
            ["vm12", 20, datetime(year=2023, month=1, day=2)],
            ["vm13", 6, datetime(year=2023, month=1, day=2)],
            ["vm14", 30, datetime(year=2023, month=1, day=2)],
            ["vm6", 0, datetime(year=2023, month=1, day=3)],
            ["vm7", 5, datetime(year=2023, month=1, day=3)],
            ["vm8", 30, datetime(year=2023, month=1, day=3)],
            ["vm9", 400, datetime(year=2023, month=1, day=3)],
            ["vm9", 40, datetime(year=2023, month=1, day=4)],
        ]

        result_df = pd.DataFrame(result, columns=["dstVmId", "cardinalityCount", "timestamp"])
        self._mock_flow_service.group_by = lambda _query: result_df

        today = datetime.combine(date.today(), datetime.min.time())
        missing_intervals = [
            Interval(today - timedelta(days=1), today),
            Interval(today - timedelta(days=2), today - timedelta(days=1)),
        ]
        partial_baselines = [self.detector._generate_baseline("site 1", interval) for interval in missing_intervals]

        self.assertEqual(2, len(partial_baselines))

        partial_baseline = partial_baselines[0]
        baseline_data = pd.read_parquet(partial_baseline.data)

        self.assertEqual(15, len(baseline_data.index))

    def test_detect_events(self):
        user_config = UserConfigInfo(
            site_id="site 1",
            name="VERTICAL_PORT_SCAN",
            org_id="org 1",
            version="1",
            last_modified_time=0,
            enabled=True,
            silent=False,
            source="source",
            block=False,
            block_confidence_threshold=50,
            mitre_tactics=[],
            mitre_techniques=[],
            settings=DetectorSettings([]),
            exclusion_settings=AnomalyExclusionSettings([]),
            run_status=DetectorRunStatus(DetectorStatusLabel.NOT_STARTED),
        )

        result = [["vm1", 0, datetime(year=2023, month=1, day=1)]]

        result_df = pd.DataFrame(result, columns=["dstVmId", "cardinalityCount", "timestamp"])
        self._mock_flow_service.group_by = lambda _query: result_df

        df = [
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
            [timedelta(days=30), np.nan, 14, 2],
        ]
        index = ["vm1", "vm2", "vm3", "vm4", "vm5", "vm6", "vm7", "vm8", "vm9", "vm10"]
        columns = ["duration", "outlier_datetimes", "mean", "std"]

        computes_detecting_on = pd.DataFrame(df, index, columns=columns)
        baseline = io.BytesIO()
        computes_detecting_on.to_parquet(baseline)

        result = [_ for _ in self.detector._detect_events(user_config, baseline, None)]

        self.assertEqual(0, len(result))


def mock_create_event_query(query: FlowGroupByQuery) -> pd.DataFrame:
    beginning_of_today = datetime.combine(date.today(), datetime.min.time())

    src_entities = [
        ["vm100", 22, beginning_of_today],
        ["vm200", 80, beginning_of_today],
        ["vm200", 22, beginning_of_today],
        ["vm200", 9000, beginning_of_today],
        ["vm200", 443, beginning_of_today],
        ["vm300", 22, beginning_of_today],
        ["vm400", 22, beginning_of_today],
    ]
    src_entities_df = pd.DataFrame(src_entities, columns=["srcVmId", "dstPort", "timestamp"])

    event_detail_result = [
        ["vm1", 130, datetime.now()],
        ["vm2", 400, datetime.now()],
        ["vm3", 28, datetime.now()],
        ["vm4", 33, datetime.now()],
        ["vm5", 5, datetime.now()],
        ["vm6", 28, datetime.now()],
        ["vm7", 99, datetime.now()],
        ["vm8", 84, datetime.now()],
        ["vm9", 10, datetime.now()],
        ["vm10", 1, datetime.now()],
    ]
    event_detail_df = pd.DataFrame(event_detail_result, columns=["dstVmId", "cardinalityCount", "timestamp"])

    if query.dimensions == ["srcVmId", "dstPort"]:
        return src_entities_df
    if query.dimensions == ["dstVmId"]:
        return event_detail_df
    raise Exception("event query with given dimensions not supported")


def mock_unique_ports_flow_query(query: FlowGroupByQuery) -> pd.DataFrame:
    vm_to_unique_ports_including_today = {"vm1": 120, "vm2": 120, "vm3": 100}

    vm_to_unique_ports_excluding_today = {"vm1": 100, "vm2": 100, "vm3": 100}

    including_today_dfs = {}
    excluding_today_dfs = {}

    beginning_of_today = datetime.combine(date.today(), datetime.min.time())

    for key, value in vm_to_unique_ports_including_today.items():
        df = pd.DataFrame([[key, value, beginning_of_today]], columns=["dstVmId", "cardinalityCount", "timestamp"])
        including_today_dfs[key] = df

    for key, value in vm_to_unique_ports_excluding_today.items():
        df = pd.DataFrame([[key, value, beginning_of_today]], columns=["dstVmId", "cardinalityCount", "timestamp"])
        excluding_today_dfs[key] = df

    query_filter = query.filter
    if query_filter is None:
        raise Exception("no filter not supported")

    outer_filter = query_filter.filter
    vm_id = None
    for filter in outer_filter["filter"]["fields"]:
        inner_filter = filter.filter  # type: ignore
        if inner_filter["filter"]["type"] == "in":
            vm_id = inner_filter["filter"]["values"][0]
            break
    if vm_id is None:
        raise Exception("couldn't find vm ID in the filter")

    if vm_id == "vm1":
        if query.intervals[0].end_time == beginning_of_today:
            return excluding_today_dfs["vm1"]
        return including_today_dfs["vm1"]

    if vm_id == "vm2":
        if query.intervals[0].end_time == beginning_of_today:
            return excluding_today_dfs["vm1"]
        return including_today_dfs["vm1"]

    if vm_id == "vm3":
        if query.intervals[0].end_time == beginning_of_today:
            return excluding_today_dfs["vm1"]
        return including_today_dfs["vm1"]

    raise Exception("vm not supported")


def test_get_detector_status(self):
    self.mock_baseline_service = Mock(BaselineService())
    self.mock_baseline_service.get_max_baseline_duration_days = lambda _x, _y, _z: 2
    original_baseline_service = self.detector._baseline_service
    self.detector._baseline_service = self.mock_baseline_service

    # Not enough baseline case
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.SUCCESS, "123", 1234, "success")
    self.assertEqual(DetectorStatusLabel.NOT_ENOUGH_BASELINE, detector_run_status.status_label)

    # FAILURE case
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.FAILURE, "123", 1234, "failed")
    self.assertEqual(DetectorStatusLabel.FAILURE, detector_run_status.status_label)

    # SUCCESS case - baseline size > min required baseline
    self.mock_baseline_service.get_max_baseline_duration_days = lambda _x, _y, _z: 7
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.SUCCESS, "123", 1234, "success")
    self.assertEqual(DetectorStatusLabel.SUCCESS, detector_run_status.status_label)

    self.detector._baseline_service = original_baseline_service
